@@ -239,6 +239,11 @@ class LibWacom:
             args=(c_void_p, c_char),
             return_type=c_int,
         ),
+        _Api(
+            name="libwacom_get_button_modeswitch_mode",
+            args=(c_void_p, c_char),
+            return_type=c_int,
+        ),
         _Api(
             name="libwacom_stylus_get_for_id",
             args=(c_void_p, c_int),
@@ -396,6 +401,11 @@ class LibWacom:
         _Enum(name="WACOM_STATUS_LED_TOUCHSTRIP2", value=4),
         _Enum(name="WACOM_STATUS_LED_DIAL", value=1),
         _Enum(name="WACOM_STATUS_LED_DIAL2", value=2),
+        _Enum(name="WACOM_MODE_SWITCH_NEXT", value=-1),
+        _Enum(name="WACOM_MODE_SWITCH_0", value=0),
+        _Enum(name="WACOM_MODE_SWITCH_1", value=1),
+        _Enum(name="WACOM_MODE_SWITCH_2", value=2),
+        _Enum(name="WACOM_MODE_SWITCH_3", value=3),
     ]
 
 
@@ -557,6 +567,15 @@ class WacomEraserType(enum.IntEnum):
     BUTTON = 3
 
 
+class WacomAxisType(enum.IntEnum):
+    NONE = 0
+    TILT = (1 << 1,)
+    ROTATION_Z = (1 << 2,)
+    DISTANCE = (1 << 3,)
+    PRESSURE = (1 << 4,)
+    SLIDER = 1 << 5
+
+
 class WacomStylus:
     class Generic(enum.IntEnum):
         PEN_WITH_ERASER = 0xFFFFF
@@ -574,7 +593,12 @@ class WacomStylus:
         for api in lib._api_prototypes:
             allowlist = ["stylus"]
             if any(api.basename.startswith(n) for n in allowlist):
-                denylist = ["stylus_get_paired_styli", "stylus_is_eraser"]
+                denylist = [
+                    "stylus_get_paired_styli",
+                    "stylus_is_eraser",
+                    "stylus_has_lens",
+                    "stylus_has_wheel",
+                ]
                 if all(not api.basename.startswith(n) for n in denylist):
                     func = getattr(lib, api.basename)
                     setattr(self, api.basename.removeprefix("stylus_"), wrapper(func))
@@ -595,18 +619,33 @@ class WacomStylus:
     def vendor_id(self) -> int:
         return self.get_vendor_id()
 
+    @property
+    def axes(self) -> list[WacomAxisType]:
+        axes = self.get_axes()
+        return [t for t in WacomAxisType if axes & t]
+
     @property
     def num_buttons(self) -> int:
         return self.get_num_buttons()
 
+    @property
+    def has_lens(self) -> bool:
+        lib = LibWacom.instance()
+        return lib.stylus_has_lens(self.stylus) != 0
+
+    @property
+    def has_wheel(self) -> bool:
+        lib = LibWacom.instance()
+        return lib.stylus_has_wheel(self.stylus) != 0
+
     @property
     def is_eraser(self) -> bool:
         lib = LibWacom.instance()
         return lib.stylus_is_eraser(self.stylus) != 0
 
     @property
-    def stylus_type(self) -> WacomEraserType:
-        return WacomEraserType(self.get_eraser_type())
+    def stylus_type(self) -> WacomStylusType:
+        return WacomStylusType(self.get_type())
 
     @property
     def eraser_type(self) -> WacomEraserType:
@@ -667,6 +706,13 @@ class WacomDevice:
                 WacomDevice.ButtonFlags.DIAL2_MODESWITCH,
             ]
 
+    class ModeSwitch(enum.IntEnum):
+        NEXT = -1
+        MODE_0 = 0
+        MODE_1 = 1
+        MODE_2 = 2
+        MODE_3 = 3
+
     def __init__(self, device, destroy=True):
         self.device = device
         self._destroy_on_del = destroy
@@ -812,6 +858,10 @@ class WacomDevice:
     def button_evdev_code(self, button: str) -> int:
         return self.get_button_evdev_code(button.encode("utf-8"))
 
+    def button_modeswitch_mode(self, button: str) -> ModeSwitch:
+        mode = self.get_button_modeswitch_mode(button.encode("utf-8"))
+        return WacomDevice.ModeSwitch(mode)
+
     def button_led_group(self, button: str) -> List[ButtonFlags]:
         return self.get_button_led_group(button.encode("utf-8"))
 
