@@ -8,8 +8,10 @@ from dataclasses import dataclass, field
 import ctypes
 import logging
 import pytest
+import string
 
 from . import (
+    WacomAxisType,
     WacomBuilder,
     WacomBustype,
     WacomDatabase,
@@ -17,6 +19,7 @@ from . import (
     WacomEraserType,
     WacomStatusLed,
     WacomStylus,
+    WacomStylusType,
 )
 
 logger = logging.getLogger(__name__)
@@ -32,6 +35,9 @@ class StylusEntry:
     axes: str = "Tilt;Pressure;Distance;"
     stylus_type: str = "General"
     eraser_type: str | None = None
+    alias_of: str | None = None
+    has_wheel: str | None = None
+    has_lens: str | None = None
 
     def add_to_config(self, config: ConfigParser):
         c = {
@@ -42,22 +48,26 @@ class StylusEntry:
             "Axes": self.axes,
             "Type": self.stylus_type,
         }
-        if self.eraser_type is not None:
-            c["EraserType"] = self.eraser_type
+        for attr in ["eraser_type", "alias_of", "has_wheel", "has_lens"]:
+            if (val := getattr(self, attr, None)) is not None:
+                key = attr.title().replace("_", "")
+                # We need lowercase true/false
+                c[key] = val if not isinstance(val, bool) else str(val).lower()
+
         config[self.id] = c
 
     @classmethod
     def generic_pen(cls) -> "StylusEntry":
         return cls(
-            id="0x0:0xfffff", name="General Pen", paired_stylus_ids=["0x0:0xffffe"]
+            id="0x0:0xaffff", name="General Pen", paired_stylus_ids=["0x0:0xafffe"]
         )
 
     @classmethod
     def generic_eraser(cls) -> "StylusEntry":
         return StylusEntry(
-            id="0x0:0xffffe",
+            id="0x0:0xafffe",
             name="General Pen Eraser",
-            paired_stylus_ids=["0x0:0xfffff"],
+            paired_stylus_ids=["0x0:0xaffff"],
             eraser_type="Invert",
         )
 
@@ -85,6 +95,11 @@ class StylusFile:
         with open(dir / filename, "w") as fd:
             config.write(fd, space_around_delimiters=False)
 
+        if logger.getEffectiveLevel() == logging.DEBUG:
+            logger.debug(f"{dir}/{filename}:")
+            for line in open(dir / filename).readlines():
+                logger.debug(f"  {line.rstrip()}")
+
 
 @dataclass
 class TabletFile:
@@ -131,6 +146,11 @@ class TabletFile:
         with open(filename, "w") as fd:
             config.write(fd, space_around_delimiters=False)
 
+        if logger.getEffectiveLevel() == logging.DEBUG:
+            logger.debug(f"{filename}:")
+            for line in open(filename).readlines():
+                logger.debug(f"  {line.rstrip()}")
+
 
 @pytest.fixture()
 def custom_datadir(tmp_path):
@@ -226,6 +246,11 @@ def test_cintiq24hdt(db):
     assert match.product_id == 0xF6
     assert match.bustype == device.BUSTYPE_USB
 
+    modes = {"A": 0, "B": 1, "C": 2, "I": 0, "J": 1, "K": 2}
+    for btn in string.ascii_uppercase[: device.num_buttons]:
+        expected_mode = modes.get(btn, WacomDevice.ModeSwitch.NEXT)
+        assert device.button_modeswitch_mode(btn) == expected_mode
+
 
 def test_cintiq13hd(db):
     libevdev = pytest.importorskip("libevdev")
@@ -249,6 +274,9 @@ def test_cintiqpro13(db):
     assert device is not None
     assert device.num_keys == 5
 
+    for btn in string.ascii_uppercase[: device.num_buttons]:
+        assert device.button_modeswitch_mode(btn) == WacomDevice.ModeSwitch.NEXT
+
 
 def test_dell_canvas(db):
     device = db.new_from_name("Dell Canvas 27")
@@ -283,6 +311,16 @@ def test_isdv4_4800(db):
     assert device.num_buttons == 0
 
 
+def test_mobilestudio_pro_modeswitch(db):
+    device = db.new_from_name("Wacom MobileStudio Pro 13")
+    assert device is not None
+
+    modes = {"H": 0, "I": 1, "J": 2, "K": 3}
+    for btn in string.ascii_uppercase[: device.num_buttons]:
+        expected_mode = modes.get(btn, WacomDevice.ModeSwitch.NEXT)
+        assert device.button_modeswitch_mode(btn) == expected_mode
+
+
 @pytest.mark.parametrize(
     "usbid,expected",
     [
@@ -514,11 +552,11 @@ def test_new_unknown_device_with_fallback(custom_datadir, fallback, bustype):
         assert device is None
 
 
-def create_uinput(name, vid, pid):
+def create_uinput(name, vid, pid, bustype=0x3):
     libevdev = pytest.importorskip("libevdev")
     dev = libevdev.Device()
     dev.name = name
-    dev.id = {"bustype": 0x3, "vendor": vid, "product": pid}
+    dev.id = {"bustype": bustype, "vendor": vid, "product": pid}
     dev.enable(
         libevdev.EV_ABS.ABS_X,
         libevdev.InputAbsInfo(minimum=0, maximum=10000, resolution=200),
@@ -553,14 +591,22 @@ def test_new_from_path_known_device(db, fallback):
     assert dev.product_id == pid
 
 
+@pytest.mark.parametrize("bustype", WacomBustype)
 @pytest.mark.parametrize(
     "fallback", (WacomDatabase.Fallback.NONE, WacomDatabase.Fallback.GENERIC)
 )
-def test_new_from_path_unknown_device(db, fallback):
+def test_new_from_path_unknown_device(db, fallback, bustype):
+    bus = {
+        WacomBustype.UNKNOWN: 0,
+        WacomBustype.USB: 0x3,
+        WacomBustype.BLUETOOTH: 0x5,
+        WacomBustype.I2C: 0x18,
+    }[bustype]
+
     name = "Unknown device"
     vid = 0x1234
     pid = 0xABAC
-    uinput = create_uinput(name, vid, pid)
+    uinput = create_uinput(name, vid, pid, bustype=bus)
 
     dev = db.new_from_path(
         uinput.devnode, fallback=fallback
@@ -572,19 +618,17 @@ def test_new_from_path_unknown_device(db, fallback):
         assert dev.name == name
         assert dev.vendor_id == 0
         assert dev.product_id == 0
+        assert dev.bustype == 0  # fallback device is always bustype 0
 
 
 @pytest.mark.parametrize(
     "feature",
-    ("Ring", "Touchstrip", "Dial"),
+    ("Ring", "Strip", "Dial"),
 )
 @pytest.mark.parametrize("count", (1, 2))
 def test_button_modeswitch(custom_datadir, feature, count):
     USBID = (0x1234, 0x5678)
 
-    # sigh, Touchstrip but StripsNumModes...
-    num_mode_key = f"{feature}s" if feature != "Touchstrip" else "Strip"
-
     extra = {
         "Buttons": {
             "Left": "A;B;C;",
@@ -592,7 +636,7 @@ def test_button_modeswitch(custom_datadir, feature, count):
             feature: "A",
         },
         "Features": {
-            num_mode_key: 4,
+            f"{feature}s": 4,
         },
     }
     if count > 1:
@@ -611,7 +655,7 @@ def test_button_modeswitch(custom_datadir, feature, count):
 
     expected_flag = {
         "Ring": WacomDevice.ButtonFlags.RING_MODESWITCH,
-        "Touchstrip": WacomDevice.ButtonFlags.TOUCHSTRIP_MODESWITCH,
+        "Strip": WacomDevice.ButtonFlags.TOUCHSTRIP_MODESWITCH,
         "Dial": WacomDevice.ButtonFlags.DIAL_MODESWITCH,
     }[feature]
 
@@ -624,7 +668,7 @@ def test_button_modeswitch(custom_datadir, feature, count):
 
     expected_flag = {
         "Ring": WacomDevice.ButtonFlags.RING2_MODESWITCH,
-        "Touchstrip": WacomDevice.ButtonFlags.TOUCHSTRIP2_MODESWITCH,
+        "Strip": WacomDevice.ButtonFlags.TOUCHSTRIP2_MODESWITCH,
         "Dial": WacomDevice.ButtonFlags.DIAL2_MODESWITCH,
     }[feature]
 
@@ -641,15 +685,12 @@ def test_button_modeswitch(custom_datadir, feature, count):
 
 @pytest.mark.parametrize(
     "feature",
-    ("Ring", "Touchstrip", "Dial"),
+    ("Ring", "Strip", "Dial"),
 )
 @pytest.mark.parametrize("count", (1, 2))
 def test_status_leds(custom_datadir, feature, count):
     USBID = (0x1234, 0x5678)
 
-    # sigh, Touchstrip but StripsNumModes...
-    num_mode_key = f"{feature}s" if feature != "Touchstrip" else "Strip"
-
     extra = {
         "Buttons": {
             "Left": "A;B;C;",
@@ -658,7 +699,7 @@ def test_status_leds(custom_datadir, feature, count):
         },
         "Features": {
             "StatusLEDs": f"{feature};{feature}2" if count > 1 else f"{feature}",
-            num_mode_key: 4,
+            f"{feature}s": 4,
         },
     }
     if count > 1:
@@ -678,7 +719,7 @@ def test_status_leds(custom_datadir, feature, count):
     expected = [
         {
             "Ring": WacomStatusLed.RING,
-            "Touchstrip": WacomStatusLed.TOUCHSTRIP,
+            "Strip": WacomStatusLed.TOUCHSTRIP,
             "Dial": WacomStatusLed.DIAL,
         }[feature]
     ]
@@ -687,7 +728,7 @@ def test_status_leds(custom_datadir, feature, count):
         expected.append(
             {
                 "Ring": WacomStatusLed.RING2,
-                "Touchstrip": WacomStatusLed.TOUCHSTRIP2,
+                "Strip": WacomStatusLed.TOUCHSTRIP2,
                 "Dial": WacomStatusLed.DIAL2,
             }[feature]
         )
@@ -778,8 +819,8 @@ def test_nonwacom_stylus_ids(tmp_path):
     assert len(styli) == 3  # 1 non-wacom, 2 generic ones
     # Order of styli is undefined
     assert sum(s.vendor_id == 0x1234 and s.tool_id == 0x9876 for s in styli) == 1
-    assert sum(s.vendor_id == 0 and s.tool_id == 0xFFFFE for s in styli) == 1
-    assert sum(s.vendor_id == 0 and s.tool_id == 0xFFFFF for s in styli) == 1
+    assert sum(s.vendor_id == 0 and s.tool_id == 0xAFFFE for s in styli) == 1
+    assert sum(s.vendor_id == 0 and s.tool_id == 0xAFFFF for s in styli) == 1
 
 
 def test_load_xdg_config_home(monkeypatch, tmp_path, custom_datadir):
@@ -801,3 +842,155 @@ def test_load_xdg_config_home(monkeypatch, tmp_path, custom_datadir):
     builder = WacomBuilder.create(usbid=usbid)
     device = db.new_from_builder(builder)
     assert device is not None and device.name == "XDGTablet"
+
+
+@pytest.mark.parametrize(
+    "override,value",
+    (
+        ("axes", "Tilt;"),
+        ("buttons", 3),
+        ("has_lens", True),
+        ("has_wheel", True),
+        ("stylus_type", WacomStylusType.AIRBRUSH),
+        ("eraser_type", WacomEraserType.INVERT),
+    ),
+)
+def test_alias_of(custom_datadir, override, value):
+    usbid = (0x1234, 0x5678)
+    matches = [f"usb|{usbid[0]:04x}|{usbid[1]:04x}"]
+    TabletFile(name="XDGTablet", matches=matches, styli=["@happy-aliases"]).write_to(
+        custom_datadir / "uniq.tablet"
+    )
+
+    stylusfile = StylusFile.default()
+    stylusfile.entries.append(
+        StylusEntry(
+            id="0x1234:0xabcd",
+            name="To be aliased",
+            group="happy-aliases",
+        )
+    )
+
+    alias = StylusEntry(
+        id="0x1234:0xffff",
+        name="AliasTester",
+        alias_of="0x1234:0xabcd",
+        group="happy-aliases",
+    )
+    if override == "stylus_type":
+        setattr(alias, override, "Airbrush")
+    elif override == "eraser_type":
+        setattr(alias, override, "Invert")
+    else:
+        setattr(alias, override, value)
+
+    stylusfile.entries.append(alias)
+    stylusfile.write_to_dir(custom_datadir)
+
+    db = WacomDatabase(path=custom_datadir)
+    builder = WacomBuilder.create(usbid=usbid)
+    device = db.new_from_builder(builder)
+    assert device is not None and device.name == "XDGTablet"
+
+    styli = device.get_styli()
+    assert len(styli) == 2
+
+    original = styli[0]
+    aliasing = styli[1]
+
+    assert original.vendor_id == 0x1234
+    assert original.tool_id == 0xABCD
+    assert original.num_buttons == 2
+    assert original.axes == [
+        WacomAxisType.TILT,
+        WacomAxisType.DISTANCE,
+        WacomAxisType.PRESSURE,
+    ]
+    assert not original.has_lens
+    assert not original.has_wheel
+    assert original.eraser_type == WacomEraserType.NONE
+
+    assert aliasing.vendor_id == 0x1234
+    assert aliasing.tool_id == 0xFFFF
+    if override == "buttons":
+        assert aliasing.num_buttons == value
+    else:
+        assert aliasing.num_buttons == 2
+
+    if override == "axes":
+        assert aliasing.axes == [WacomAxisType.TILT]
+    else:
+        assert aliasing.axes == [
+            WacomAxisType.TILT,
+            WacomAxisType.DISTANCE,
+            WacomAxisType.PRESSURE,
+        ]
+
+    if override == "stylus_type":
+        assert aliasing.stylus_type == value
+    else:
+        assert aliasing.stylus_type == WacomStylusType.GENERAL
+
+    if override == "eraser_type":
+        assert aliasing.eraser_type == value
+    else:
+        assert aliasing.eraser_type == WacomEraserType.NONE
+
+    if override == "has_lens":
+        assert aliasing.has_lens is value
+    else:
+        assert aliasing.has_lens is False
+
+    if override == "has_wheel":
+        assert aliasing.has_wheel is value
+    else:
+        assert aliasing.has_wheel is False
+
+
+@pytest.mark.parametrize("alias", ("0x1234:0xaaaa", "0xabcd", "banana"))
+def test_alias_of_invalid(custom_datadir, alias):
+    usbid = (0x1234, 0x5678)
+    matches = [f"usb|{usbid[0]:04x}|{usbid[1]:04x}"]
+    TabletFile(name="XDGTablet", matches=matches, styli=["@happy-aliases"]).write_to(
+        custom_datadir / "uniq.tablet"
+    )
+
+    stylusfile = StylusFile.default()
+    stylusfile.entries = []
+    stylusfile.entries.append(
+        StylusEntry(
+            id="0x1234:0xabcd",
+            name="To be aliased",
+            group="happy-aliases",
+        )
+    )
+
+    alias = StylusEntry(
+        id="0x1234:0xffff",
+        name="AliasTester",
+        alias_of=alias,
+        group="happy-aliases",
+    )
+    stylusfile.entries.append(alias)
+    stylusfile.write_to_dir(custom_datadir)
+
+    db = WacomDatabase(path=custom_datadir)
+    builder = WacomBuilder.create(usbid=usbid)
+    device = db.new_from_builder(builder)
+    assert device is not None and device.name == "XDGTablet"
+
+    styli = device.get_styli()
+    assert len(styli) == 1
+
+    original = styli[0]
+    assert original.vendor_id == 0x1234
+    assert original.tool_id == 0xABCD
+    assert original.num_buttons == 2
+    assert original.axes == [
+        WacomAxisType.TILT,
+        WacomAxisType.DISTANCE,
+        WacomAxisType.PRESSURE,
+    ]
+    assert not original.has_lens
+    assert not original.has_wheel
+    assert original.eraser_type == WacomEraserType.NONE
